
local addon = LibStub("AceAddon-3.0"):NewAddon("PetLeash", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
_G.PetLeash = addon

local L	= LibStub("AceLocale-3.0"):GetLocale("PetLeash")

local defaults = {
	profile = {
        enable = true,
        enableInCombat = false,
    },
    char = {
        ignore_pets = {}       -- [spellid] = true, hide.  if false, don't hide
    }
}

local function config_toggle_get(info) return addon.db.profile[info[1]] end
local function config_toggle_set(info,v) addon.db.profile[info[1]] = v end

local options = {
    name = "PetLeash",
    handler = PetLeash,
    type = 'group',
    args = {
        enable = {
            type = "toggle",
            name = ENABLE,
            desc = L["Enable Auto-Summon"], 
            order = 1,
            width = "full",
            get = function(info) return addon:IsEnabledSummoning() end,
            set = function(info,v) addon:EnableSummoning(v) end,
        },
        enableInCombat = {
            type = "toggle",
            name = L["Enable In Combat"],
            order = 2,
            width = "double",
            get = config_toggle_get,
            set = config_toggle_set
        },
        config = {
            type = "execute",
            name = L["Open Configuration"],
            dialogHidden = true,
            order = 3,
            func = function(info) addon:OpenOptions() end
        },
        pet = {
            type = "group",
            inline = true,
            name = "Enabled Pets",
            order = 4,
            cmdHidden = true,
            args = {}
        },
        resummon = {
            type = "execute",
            name = L["Summon Another Pet"],
            desc = L["Desummon your current pet and summon another pet.  Enable summoning if needed."],
            order = 10,
            func = function(info) addon:ResummonPet() end
        },
        desummon = {
            type = "execute",
            name = L["Desummon Pet"],
            desc = L["Desummon your currently summoned pet.  Disable summoning."],
            order = 11,
            func = function(info) addon:DesummonPet() end
        },
    },
}

local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")
local LDB = LibStub("LibDataBroker-1.1", true)

local delay_timer = 30  -- extra large value 

function addon:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("PetLeashDB", defaults)
    
    LibStub("AceConfig-3.0"):RegisterOptionsTable(self.name, options, {"petleash", "pl"})
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(self.name, self.name)
    self.optionsFrame.About = LibStub("LibAboutPanel").new(self.name, self.name)
    
    self.usable_pets = {}   -- spellid, spellid
    self.pet_map = {}       -- spellid -> {id,name} (complete)
    
    self:ScheduleRepeatingTimer("PeroidicCheckPet", 1.0)
    self:RegisterEvent("PLAYER_ENTERING_WORLD")
    self:RegisterEvent("COMPANION_LEARNED")
    
    self:InitBroker()
end

function addon:IsEnabledSummoning()
    return self.db.profile.enable
end

function addon:EnableSummoning(v)
    local oldv = self.db.profile.enable

    if((not oldv) ~= (not v)) then
        self.db.profile.enable = v
        
        -- TODO: is there a better way to trigger config update?
        AceConfigRegistry:NotifyChange("PetLeash")
    end
end

function addon:OpenOptions()
    InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
end

-- utility functions

local function HasCompanion(pettype)
    for id = 1,GetNumCompanions(pettype) do
        local _,_,_,_,issum = GetCompanionInfo(pettype, id)
        if(issum) then
            return id
        end
    end
    return nil
end

local function IsMounted()
    return _G.IsMounted()
    --return HasCompanion("MOUNT")
end

local function IsPetted(id)
    if(id) then
        local _,_,_,_,issum = GetCompanionInfo("CRITTER", id)
        if(issum) then
            return id
        end
    else
        return HasCompanion("CRITTER")
    end
end

local function IsCasting()
    return (UnitCastingInfo("player") or UnitChannelInfo("player"))
end

local InCombat = InCombatLockdown       -- shorthand

local function CanSummonPet()
    return
        -- are we busy?
            not IsCasting()
        and not IsMounted()
        and not UnitInVehicle("player")
        and not UnitIsGhost("player")
        and not UnitIsDead("player")
        and not UnitOnTaxi("player")
        -- verify we have pets
        and GetNumCompanions("CRITTER") > 0
        -- gcd check
        and GetCompanionCooldown("CRITTER", 1) == 0
end

-- pet list handling

function addon:LoadPets(updateconfig)
    wipe(self.usable_pets)
    
    for i = 1,GetNumCompanions("CRITTER") do
        local _, name, spellid = GetCompanionInfo("CRITTER", i)
        
        if(not self.db.char.ignore_pets[spellid]) then
            table.insert(self.usable_pets, spellid)
        end

        if(not self.pet_map[spellid]) then
            self.pet_map[spellid] = {}
        end
        self.pet_map[spellid].id = i
        self.pet_map[spellid].name = name 
    end
    
    if(updateconfig == nil or updateconfig) then
        self:UpdateConfigTables()
    end
end

function addon:UpdateConfigTables()
    local args = options.args.pet.args
    
    wipe(args)
    
    for i = 1,GetNumCompanions("CRITTER") do
        local _, name, spellid = GetCompanionInfo("CRITTER", i)
        
        args[tostring(spellid)] = {
            type = "toggle",
            name = name,
            order = 1,
            get = "Config_PetToggle_Get",
            set = "Config_PetToggle_Set"
        }
    end
    
    -- Config Tables changed!
    AceConfigRegistry:NotifyChange("PetLeash")
end

function addon:Config_PetToggle_Set(info, v)
    if(v) then
        self.db.char.ignore_pets[tonumber(info[#info])] = nil
    else
        self.db.char.ignore_pets[tonumber(info[#info])] = true
    end
    
    self:LoadPets(false)
end

function addon:Config_PetToggle_Get(info)
    return not self.db.char.ignore_pets[tonumber(info[#info])]
end

-- events

function addon:PLAYER_ENTERING_WORLD()
    delay_timer = 2       -- wait a sec before trying to summon pet after logging in
    
    self:LoadPets()
end

function addon:COMPANION_LEARNED()
    self:LoadPets()
end

function addon:PeroidicCheckPet()
    if(delay_timer > 0) then
        delay_timer = delay_timer - 1
        return
    end
    
    self:CheckPet()
end

function addon:CheckPet()
    if(not self.db.profile.enable) then
        return
    elseif(not self.db.profile.enableInCombat and InCombat()) then
        return
    end

    local pet_id = IsPetted(self.pet_id) or IsPetted()
    if(pet_id) then
        self.pet_id = pet_id
    else
        if(CanSummonPet()) then
            self:SummonPet()
        end
    end
end

function addon:SummonPet()
    delay_timer = 5       -- stall checking again for a bit

    local n = #self.usable_pets
    if(n > 0) then
        local random_spellid = self.usable_pets[math.random(n)]
        CallCompanion("CRITTER", self.pet_map[random_spellid].id)
    end
end

function addon:DesummonPet(disable)
    DismissCompanion("CRITTER")
    if(not disable or disable == nil) then
        addon:EnableSummoning(false)
    end
end

function addon:ResummonPet()
    addon:EnableSummoning(true)

    self:DesummonPet(true)
    self:SummonPet()
end

-- LDB bits

local BrokerMenu_InitalizeMenu, BrokerMenu_SummonPet, BrokerMenu_PetRangeString

function addon:InitBroker()
	if (LDB and not self.broker) then
		self.broker = LDB:NewDataObject("PetLeash", {
			label = "PetLeash",
			type = "launcher",
			icon = "Interface\\Icons\\INV_Box_PetCarrier_01",
            OnClick = function(...) self:Broker_OnClick(...) end,
            OnTooltipShow = function(...) self:Broker_OnTooltipShow(...) end
		})
	end
    
    if(not self.brokerMenu) then
        self.brokerMenu = CreateFrame("FRAME", "PetLeashBrokerMenu",
                                      UIParent, "UIDropDownMenuTemplate")
        UIDropDownMenu_Initialize(self.brokerMenu, BrokerMenu_InitalizeMenu, "MENU")
    end
end

function addon:Broker_OnTooltipShow(tt)
    tt:AddLine("PetLeash")
    tt:AddLine(" ")
    tt:AddDoubleLine(("|cffeeeeee%s|r "):format(L["Auto Summon:"]),
                    self:IsEnabledSummoning() and ("|cff00ff00%s|r"):format(L["Enabled"])
                                              or ("|cffff0000%s|r"):format(L["Disabled"]))
    tt:AddLine(" ")
    tt:AddLine(("|cff69b950%s|r |cffeeeeee%s|r"):format(L["Left-Click:"], L["Toggle Non-Combat Pet"]))
    tt:AddLine(("|cff69b950%s|r |cffeeeeee%s|r"):format(L["Right-Click:"], L["Pet Menu"]))
    tt:AddLine(("|cff69b950%s|r |cffeeeeee%s|r"):format(L["Ctrl + Click:"], L["Open Configuration Panel"]))
end

function addon:Broker_OnClick(frame, button)
    if(IsControlKeyDown()) then
        self:OpenOptions()
    elseif(button == "LeftButton") then
        if(IsPetted()) then
            self:DesummonPet()
        else
            self:ResummonPet()
        end
    else
        ToggleDropDownMenu(1, nil, self.brokerMenu, frame, 0, 0)
    end
end

local function iter_utf8(s)
    -- Src: http://lua-users.org/wiki/LuaUnicode
    return string.gmatch(s, "([%z\1-\127\194-\244][\128-\191]*)")
end

local function str_range_diff(a,b)
    if(not b) then return iter_utf8(a)() end
    if(not a) then return nil, iter_utf8(b)() end
   
    local iter_a = iter_utf8(a)
    local iter_b = iter_utf8(b)
    local char_a, char_b = iter_a(), iter_b()
    local r = ""
    while char_a and char_b do
        if(char_a ~= char_b) then
            return r..char_a, r..char_b
        end

        r = r .. char_a

        char_a = iter_a()
        char_b = iter_b()
    end

    return r
end

local function safe_GetCritterName(id)
    if(id <= 0 or id > GetNumCompanions("CRITTER")) then
        return nil
    end
    return (select(2, GetCompanionInfo("CRITTER", id)))
end

-- for two pet ids, forming a span from "a" to "b", generate a string
-- representing this span.  For example:  A - Z
function BrokerMenu_PetRangeString(a, b)
    local _, part_a = str_range_diff(safe_GetCritterName(a-1), safe_GetCritterName(a))
    local part_b = str_range_diff(safe_GetCritterName(b), safe_GetCritterName(b+1))
    
    return string.format("%s - %s", part_a, part_b)
end

function BrokerMenu_InitalizeMenu(frame,level, menuList)
    local ncritters  = GetNumCompanions("CRITTER")
    
    -- If we have more than 25 critters, then split into equal
    -- groups of no more than 25 each
    if(not level or level == 1) then
        if(ncritters == 0) then
            -- Nothing to do!
        elseif(ncritters <= 25) then
            -- don't have to split :)
            local info = UIDropDownMenu_CreateInfo()
            info.text = L["Pets"]
            info.notCheckable = true
            info.hasArrow = true
            info.menuList = "1-"..ncritters
            UIDropDownMenu_AddButton(info, level) 
        else
            -- Split
            local numlines = math.ceil(ncritters/25)
            local generic_linesz = ncritters/numlines       -- average size
            
            for line = 1,numlines do
                -- start to finish are inclusive
                local start = math.floor(generic_linesz*(line-1)+1)
                local finish = math.floor(generic_linesz*line)
                local linesize = finish-start+1
            
                local info = UIDropDownMenu_CreateInfo()
                info.text = string.format(L["Pets"].." (%s)",BrokerMenu_PetRangeString(start, finish))
                info.notCheckable = true
                info.hasArrow = true
                info.menuList = string.format("%d-%d", start, finish)
                UIDropDownMenu_AddButton(info, level) 
            end
        end
    elseif(level == 2) then
        local start,finish = strsplit("-",menuList)     -- decode
        for i = tonumber(start),tonumber(finish) do
            local _, name = GetCompanionInfo("CRITTER", i)
        
            local info = UIDropDownMenu_CreateInfo()
            info.text = name
            info.value = i
            info.notCheckable = true
            info.func = BrokerMenu_SummonPet

            UIDropDownMenu_AddButton(info, level)            
        end
    end
end

function BrokerMenu_SummonPet(info)
    CallCompanion("CRITTER", info.value)
end
